"""cleanup migration for consent multitranslation.
Removes deprecated fields from privacy notice and privacy experience related tables and adds
some constraints after the previous data migration guarantees

Revision ID: 32497f08e227
Revises: 14acee6f5459
Create Date: 2024-03-05 17:08:07.093568

"""

from datetime import datetime

import sqlalchemy as sa
from alembic import op
from loguru import logger
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "32497f08e227"
down_revision = "14acee6f5459"
branch_labels = None
depends_on = None


def upgrade():
    logger.info(
        f"Starting Consent Multitranslation Schema Migration (#3) {datetime.now()}"
    )

    # Privacy Experience > experience_config_id is guaranteed to exist after the data migration so we can make this non-nullable
    op.alter_column(
        "privacyexperience",
        "experience_config_id",
        existing_type=sa.VARCHAR(),
        nullable=False,
    )
    # Resetting Experience Config name back to non-nullable after data migration
    op.alter_column(
        "privacyexperienceconfig", "name", existing_type=sa.VARCHAR(), nullable=False
    )
    # Resetting PrivacyNoticeHistory title back to non-nullable after data migration
    op.alter_column(
        "privacynoticehistory",
        "title",
        existing_type=sa.VARCHAR(),
        nullable=False,
    )
    # Dropping the index as a prereq before removing the column
    op.drop_index(
        "ix_privacyexperienceconfig_banner_enabled",
        table_name="privacyexperienceconfig",
    )
    # Removing the "translatable" columns that have been moved to ExperienceTranslation
    op.drop_column("privacyexperienceconfig", "privacy_policy_url")
    op.drop_column("privacyexperienceconfig", "acknowledge_button_label")
    op.drop_column("privacyexperienceconfig", "description")
    op.drop_column("privacyexperienceconfig", "banner_enabled")
    op.drop_column("privacyexperienceconfig", "banner_description")
    op.drop_column("privacyexperienceconfig", "accept_button_label")
    op.drop_column("privacyexperienceconfig", "privacy_preferences_link_label")
    # The Experience Config no longer stores a version
    op.drop_column("privacyexperienceconfig", "version")
    op.drop_column("privacyexperienceconfig", "banner_title")
    op.drop_column("privacyexperienceconfig", "title")
    op.drop_column("privacyexperienceconfig", "privacy_policy_link_label")
    # New concept by the same name is now stored on the ExperienceTranslatoin table
    op.drop_column("privacyexperienceconfig", "is_default")
    op.drop_column("privacyexperienceconfig", "save_button_label")
    op.drop_column("privacyexperienceconfig", "reject_button_label")
    # The historical config no longer references the current config, but a translation instead
    op.drop_column("privacyexperienceconfighistory", "experience_config_id")
    # Dropping notice regions index as a prereq for deleting the column
    op.drop_index("ix_privacynotice_regions", table_name="privacynotice")
    # Deleting notice constructs that have been moved to the Experience side
    op.drop_column("privacynotice", "displayed_in_api")
    op.drop_column("privacynotice", "description")
    op.drop_column("privacynotice", "regions")
    op.drop_column("privacynotice", "displayed_in_overlay")
    op.drop_column("privacynotice", "displayed_in_privacy_center")
    op.drop_column("privacynotice", "version")
    # Dropping the notice history > notice id FK as a prereq for dropping the column
    op.drop_constraint(
        "privacynoticehistory_privacy_notice_id_fkey",
        "privacynoticehistory",
        type_="foreignkey",
    )
    op.drop_column("privacynoticehistory", "privacy_notice_id")
    # Similarly removing deprecated concepts off the notice templates
    op.drop_index(
        "ix_privacynoticetemplate_regions", table_name="privacynoticetemplate"
    )
    op.drop_column("privacynoticetemplate", "displayed_in_api")
    op.drop_column("privacynoticetemplate", "description")
    op.drop_column("privacynoticetemplate", "regions")
    op.drop_column("privacynoticetemplate", "displayed_in_overlay")
    op.drop_column("privacynoticetemplate", "displayed_in_privacy_center")
    # Removing the reference from privacy preference history and served notice history to the experience
    op.drop_column("privacypreferencehistory", "privacy_experience_id")
    op.drop_column("servednoticehistory", "privacy_experience_id")

    logger.info(f"Consent Multitranslation Migrations Complete {datetime.now()}")


def downgrade():
    """Downgrade edited to make all columns nullable"""
    op.add_column(
        "servednoticehistory",
        sa.Column(
            "privacy_experience_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column(
            "privacy_experience_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column(
            "displayed_in_privacy_center",
            sa.BOOLEAN(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column(
            "displayed_in_overlay", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column(
            "regions",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column("description", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacynoticetemplate",
        sa.Column("displayed_in_api", sa.BOOLEAN(), autoincrement=False, nullable=True),
    )
    op.create_index(
        "ix_privacynoticetemplate_regions",
        "privacynoticetemplate",
        ["regions"],
        unique=False,
    )
    op.add_column(
        "privacynoticehistory",
        sa.Column(
            "privacy_notice_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.create_foreign_key(
        "privacynoticehistory_privacy_notice_id_fkey",
        "privacynoticehistory",
        "privacynotice",
        ["privacy_notice_id"],
        ["id"],
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "displayed_in_privacy_center",
            sa.BOOLEAN(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "version",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "displayed_in_overlay", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column(
            "regions",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacynotice",
        sa.Column("description", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacynotice",
        sa.Column("displayed_in_api", sa.BOOLEAN(), autoincrement=False, nullable=True),
    )
    op.create_index(
        "ix_privacynotice_regions", "privacynotice", ["regions"], unique=False
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "experience_config_id", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.create_foreign_key(
        "privacyexperienceconfighistory_experience_config_id_fkey",
        "privacyexperienceconfighistory",
        "privacyexperienceconfig",
        ["experience_config_id"],
        ["id"],
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "reject_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "save_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("is_default", sa.BOOLEAN(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "privacy_policy_link_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "version",
            postgresql.DOUBLE_PRECISION(precision=53),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "privacy_preferences_link_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "accept_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_enabled", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("description", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "acknowledge_button_label", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "privacy_policy_url", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.create_index(
        "ix_privacyexperienceconfig_banner_enabled",
        "privacyexperienceconfig",
        ["banner_enabled"],
        unique=False,
    )
    op.alter_column(
        "privacyexperienceconfig", "name", existing_type=sa.VARCHAR(), nullable=True
    )
    op.alter_column(
        "privacyexperience",
        "experience_config_id",
        existing_type=sa.VARCHAR(),
        nullable=True,
    )



from __future__ import annotations

from datetime import datetime
from typing import List, Optional, Set, Tuple, Union

from fastapi import Depends, HTTPException, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy import column
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Query, Session
from starlette.responses import StreamingResponse
from starlette.status import (
    HTTP_200_OK,
    HTTP_400_BAD_REQUEST,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)

from fides.api.api.deps import get_config_proxy, get_db
from fides.api.common_exceptions import (
    IdentityVerificationException,
    RedisNotConfigured,
)
from fides.api.db.seed import DEFAULT_CONSENT_POLICY
from fides.api.models.messaging import get_messaging_method
from fides.api.models.privacy_request import (
    Consent,
    ConsentRequest,
    ProvidedIdentity,
    ProvidedIdentityType,
)
from fides.api.models.property import Property
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.messaging.messaging import MessagingMethod
from fides.api.schemas.privacy_request import BulkPostPrivacyRequests
from fides.api.schemas.privacy_request import Consent as ConsentSchema
from fides.api.schemas.privacy_request import (
    ConsentPreferences,
    ConsentPreferencesWithVerificationCode,
    ConsentReport,
    ConsentRequestCreate,
    ConsentRequestResponse,
    ConsentWithExecutableStatus,
    PrivacyRequestCreate,
    VerificationCode,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.service.deps import get_messaging_service
from fides.api.util.api_router import APIRouter
from fides.api.util.consent_util import (
    get_or_create_fides_user_device_id_provided_identity,
)
from fides.api.util.endpoint_utils import validate_start_and_end_filters
from fides.api.util.logger import Pii
from fides.common.api.scope_registry import CONSENT_READ
from fides.common.api.v1.urn_registry import (
    CONSENT_REQUEST,
    CONSENT_REQUEST_PREFERENCES,
    CONSENT_REQUEST_PREFERENCES_WITH_ID,
    CONSENT_REQUEST_VERIFY,
    V1_URL_PREFIX,
)
from fides.config import CONFIG
from fides.config.config_proxy import ConfigProxy
from fides.service.messaging.messaging_service import MessagingService
from fides.service.privacy_request.privacy_request_service import PrivacyRequestService

router = APIRouter(tags=["Consent"], prefix=V1_URL_PREFIX)


CONFIG_JSON_PATH = "clients/privacy-center/config/config.json"


def _filter_consent(
    db: Session,
    query: Query,
    data_use: Optional[str] = None,
    has_gpc_flag: Optional[bool] = None,
    opt_in: Optional[bool] = None,
    created_lt: Optional[datetime] = None,
    created_gt: Optional[datetime] = None,
    updated_lt: Optional[datetime] = None,
    updated_gt: Optional[datetime] = None,
    identity: Optional[str] = None,
) -> Query:
    """Filter Consent records against the params passed in."""
    validate_start_and_end_filters(
        [
            (created_lt, created_gt, "created"),
            (updated_lt, updated_gt, "updated"),
        ]
    )

    if identity:
        identity_hashes = ProvidedIdentity.hash_value_for_search(identity)
        identities: Set[str] = {
            identity[0]
            for identity in ProvidedIdentity.filter(
                db=db,
                conditions=(ProvidedIdentity.hashed_value.in_(identity_hashes)),
            ).values(column("id"))
        }
        query = query.filter(Consent.provided_identity_id.in_(identities))

    if data_use:
        query = query.filter(Consent.data_use == data_use)
    if has_gpc_flag is not None:
        query = query.filter(Consent.has_gpc_flag == has_gpc_flag)
    if opt_in is not None:
        query = query.filter(Consent.opt_in == opt_in)

    if created_lt:
        query = query.filter(Consent.created_at < created_lt)
    if created_gt:
        query = query.filter(Consent.created_at > created_gt)
    if updated_lt:
        query = query.filter(Consent.updated_at < updated_lt)
    if updated_gt:
        query = query.filter(Consent.updated_at > updated_gt)
    return query


@router.get(
    CONSENT_REQUEST_PREFERENCES,
    dependencies=[Security(verify_oauth_client, scopes=[CONSENT_READ])],
    status_code=HTTP_200_OK,
    response_model=Page[ConsentReport],
)
def report_consent_requests(
    *,
    db: Session = Depends(get_db),
    params: Params = Depends(),
    data_use: Optional[str] = None,
    has_gpc_flag: Optional[bool] = None,
    opt_in: Optional[bool] = None,
    created_lt: Optional[datetime] = None,
    created_gt: Optional[datetime] = None,
    updated_lt: Optional[datetime] = None,
    updated_gt: Optional[datetime] = None,
    identity: Optional[str] = None,
) -> Union[StreamingResponse, AbstractPage[ConsentReport]]:
    """Provides a paginated list of all consent requests sorted by the most recently updated."""

    query = Consent.query(db).order_by(Consent.updated_at.desc())
    query = _filter_consent(
        db,
        query,
        data_use,
        has_gpc_flag,
        opt_in,
        created_lt,
        created_gt,
        updated_lt,
        updated_gt,
        identity,
    )
    paginated = paginate(query, params)
    paginated.items = [  # type: ignore
        _prepare_consent_report(
            db=db,
            consent=item,
        )
        for item in paginated.items  # type: ignore
    ]
    return paginated


@router.post(
    CONSENT_REQUEST,
    status_code=HTTP_200_OK,
    response_model=ConsentRequestResponse,
)
def create_consent_request(
    *,
    db: Session = Depends(get_db),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
    messaging_service: MessagingService = Depends(get_messaging_service),
    data: ConsentRequestCreate,
) -> ConsentRequestResponse:
    """Creates a verification code for the user to verify access to manage consent preferences."""
    if not CONFIG.redis.enabled:
        raise RedisNotConfigured(
            "Application Redis cache required, but it is currently disabled! Please update your application configuration to enable integration with a Redis cache."
        )
    # TODO: (PROD-2142)- pass in property id here
    if data.property_id:
        valid_property: Optional[Property] = Property.get_by(
            db, field="id", value=data.property_id
        )
        if not valid_property:
            raise HTTPException(
                HTTP_400_BAD_REQUEST,
                detail="The property id provided is invalid",
            )

    identity = data.identity
    if (
        not identity.email
        and not identity.phone_number
        and not identity.fides_user_device_id
        and not identity.external_id
    ):
        raise HTTPException(
            HTTP_400_BAD_REQUEST,
            detail="An email address, phone number, fides_user_device_id, or external_id is required",
        )

    provided_identity = _get_or_create_provided_identity(
        db=db,
        identity_data=identity,
    )

    consent_request_data = {
        "provided_identity_id": provided_identity.id,
        "property_id": getattr(data, "property_id", None),
        "source": getattr(data, "source", None),
    }
    consent_request = ConsentRequest.create(db, data=consent_request_data)

    consent_request.persist_custom_privacy_request_fields(
        db=db, custom_privacy_request_fields=data.custom_privacy_request_fields
    )

    messaging_service.send_verification_code(
        consent_request, data.identity, data.property_id
    )

    return ConsentRequestResponse(
        consent_request_id=consent_request.id,
    )


@router.post(
    CONSENT_REQUEST_VERIFY,
    status_code=HTTP_200_OK,
    response_model=ConsentPreferences,
)
def consent_request_verify(
    *,
    consent_request_id: str,
    db: Session = Depends(get_db),
    data: VerificationCode,
) -> ConsentPreferences:
    """Verifies the verification code and returns the current consent preferences if successful.

    Note that this returns just Consent records - which is the old workflow that saves Consent with respect to a data use.
    """
    _, provided_identity = _get_consent_request_and_provided_identity(
        db=db,
        consent_request_id=consent_request_id,
        verification_code=data.code,
    )

    if not provided_identity.hashed_value:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail="Provided identity missing"
        )

    return _prepare_consent_preferences(db, provided_identity)


@router.get(
    CONSENT_REQUEST_PREFERENCES_WITH_ID,
    status_code=HTTP_200_OK,
    response_model=ConsentPreferences,
    responses={
        HTTP_200_OK: {
            "consent": [
                {
                    "data_use": "marketing.advertising",
                    "data_use_description": "We may use some of your personal information for advertising performance "
                    "analysis and audience modeling for ongoing advertising which may be "
                    "interpreted as 'Data Sharing' under some regulations.",
                    "opt_in": True,
                    "highlight": False,
                },
                {
                    "data_use": "functional",
                    "data_use_description": "We may use some of your personal information to collect analytics about "
                    "how you use our products & services, in order to improve our service.",
                    "opt_in": False,
                },
            ]
        },
        HTTP_404_NOT_FOUND: {"detail": "Consent request not found"},
        HTTP_400_BAD_REQUEST: {
            "detail": "Retrieving consent preferences without identity verification is "
            "only supported with disable_consent_identity_verification set to true"
        },
    },
)
def get_consent_preferences_no_id(
    *,
    db: Session = Depends(get_db),
    config_proxy: ConfigProxy = Depends(get_config_proxy),
    consent_request_id: str,
) -> ConsentPreferences:
    """Returns the current consent preferences if successful."""

    if not config_proxy.execution.disable_consent_identity_verification:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Retrieving consent preferences without identity verification is "
            "only supported with disable_consent_identity_verification set to true",
        )

    _, provided_identity = _get_consent_request_and_provided_identity(
        db=db,
        consent_request_id=consent_request_id,
        verification_code=None,
    )

    if not provided_identity.hashed_value:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail="Provided identity missing"
        )

    return _prepare_consent_preferences(db, provided_identity)


@router.post(
    CONSENT_REQUEST_PREFERENCES,
    dependencies=[Security(verify_oauth_client, scopes=[CONSENT_READ])],
    status_code=HTTP_200_OK,
    response_model=ConsentPreferences,
)
def get_consent_preferences(
    *,
    db: Session = Depends(get_db),
    data: Identity,
) -> ConsentPreferences:
    """Gets the consent preferences for the specified user."""
    if not data.email and not data.phone_number:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="No identity information provided"
        )

    # From the above check we know at least one exists
    lookup = data.email if data.email else data.phone_number

    identity = ProvidedIdentity.filter(
        db,
        conditions=(
            (
                ProvidedIdentity.hashed_value.in_(
                    ProvidedIdentity.hash_value_for_search(str(lookup))
                )
            )
            & (ProvidedIdentity.privacy_request_id.is_(None))
        ),
    ).first()

    if not identity:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Identity not found")

    return _prepare_consent_preferences(db, identity)


def queue_privacy_request_to_propagate_consent_old_workflow(
    db: Session,
    provided_identity: ProvidedIdentity,
    policy: Union[FidesKey, str],
    consent_preferences: ConsentPreferences,
    consent_request: ConsentRequest,
    executable_consents: Optional[List[ConsentWithExecutableStatus]] = [],
    browser_identity: Optional[Identity] = None,
) -> Optional[BulkPostPrivacyRequests]:
    """
    Queue a privacy request to carry out propagating consent preferences server-side to third-party systems.

    Only propagate consent preferences which are considered "executable" by the current system. If none of the
    consent preferences are executable, no Privacy Request is queued.

    # TODO Slated for deprecation
    """
    # Create an identity based on any provided browser_identity
    identity = browser_identity if browser_identity else Identity()
    setattr(
        identity,
        provided_identity.field_name,  # type:ignore[attr-defined]
        provided_identity.encrypted_value["value"],  # type:ignore[index]
    )  # Pull the information on the ProvidedIdentity for the ConsentRequest to pass along to create a PrivacyRequest

    executable_data_uses = [
        ec.data_use for ec in executable_consents or [] if ec.executable
    ]

    # Restrict consent preferences to just those that are executable
    executable_consent_preferences: List[ConsentSchema] = [
        pref
        for pref in consent_preferences.consent or []
        if pref.data_use in executable_data_uses
    ]

    if not executable_consent_preferences:
        logger.info(
            "Skipping propagating consent preferences to third-party services as "
            "specified consent preferences: {} are not executable.",
            [pref.data_use for pref in consent_preferences.consent or []],
        )
        return None

    logger.info("Executable consent options: {}", executable_data_uses)

    # It's a bit weird to initialize the privacy request service here,
    # but this logic is slated for deprecation so it's not worth doing a larger refactor
    privacy_request_service = PrivacyRequestService(
        db, ConfigProxy(db), MessagingService(db, CONFIG, ConfigProxy(db))
    )

    privacy_request_results: BulkPostPrivacyRequests = (
        privacy_request_service.create_bulk_privacy_requests(
            [
                PrivacyRequestCreate(
                    identity=identity,
                    policy_key=policy,
                    consent_preferences=executable_consent_preferences,
                    consent_request_id=consent_request.id,
                    custom_privacy_request_fields=consent_request.get_persisted_custom_privacy_request_fields(),
                    source=consent_request.source,
                )
            ],
            authenticated=True,
        )
    )

    if privacy_request_results.failed or not privacy_request_results.succeeded:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=privacy_request_results.failed[0].message,
        )

    return privacy_request_results


@router.patch(
    CONSENT_REQUEST_PREFERENCES_WITH_ID,
    status_code=HTTP_200_OK,
    response_model=ConsentPreferences,
)
def set_consent_preferences(
    *,
    consent_request_id: str,
    db: Session = Depends(get_db),
    data: ConsentPreferencesWithVerificationCode,
) -> ConsentPreferences:
    """Verifies the verification code and saves the user's consent preferences if successful.

    Note that this allows you to save Consent records under our old workflow that saves Consent with respect to a data use.

    # TODO Slated for deprecation
    """
    consent_request, provided_identity = _get_consent_request_and_provided_identity(
        db=db,
        consent_request_id=consent_request_id,
        verification_code=data.code,
    )
    consent_request.preferences = [
        schema.model_dump(mode="json") for schema in data.consent
    ]
    consent_request.save(db=db)

    if not provided_identity.hashed_value:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail="Provided identity missing"
        )

    for preference in data.consent:
        current_preference = Consent.filter(
            db=db,
            conditions=(Consent.provided_identity_id == provided_identity.id)
            & (Consent.data_use == preference.data_use),
        ).first()

        if current_preference:
            current_preference.update(db, data=dict(preference))
        else:
            preference_dict = dict(preference)
            preference_dict["provided_identity_id"] = provided_identity.id
            try:
                Consent.create(db, data=preference_dict)
            except IntegrityError as exc:
                raise HTTPException(
                    status_code=HTTP_400_BAD_REQUEST, detail=Pii(str(exc))
                )

    consent_preferences: ConsentPreferences = _prepare_consent_preferences(
        db, provided_identity
    )

    # Note: This just queues the PrivacyRequest for processing
    privacy_request_creation_results: Optional[BulkPostPrivacyRequests] = (
        queue_privacy_request_to_propagate_consent_old_workflow(
            db,
            provided_identity,
            data.policy_key or DEFAULT_CONSENT_POLICY,
            consent_preferences,
            consent_request,
            data.executable_options,
            data.browser_identity,
        )
    )

    if privacy_request_creation_results:
        consent_request.privacy_request_id = privacy_request_creation_results.succeeded[
            0
        ].id
        consent_request.save(db=db)

    return consent_preferences


def _get_or_create_provided_identity(
    db: Session,
    identity_data: Identity,
) -> ProvidedIdentity:
    """Based on target identity type, retrieves or creates associated ProvidedIdentity"""
    target_identity_type: str = infer_target_identity_type(db, identity_data)
    if target_identity_type == ProvidedIdentityType.email.value and identity_data.email:
        identity = ProvidedIdentity.filter(
            db=db,
            conditions=(
                (ProvidedIdentity.field_name == ProvidedIdentityType.email.value)
                & (
                    ProvidedIdentity.hashed_value.in_(
                        ProvidedIdentity.hash_value_for_search(identity_data.email)
                    )
                )
                & (ProvidedIdentity.privacy_request_id.is_(None))
            ),
        ).first()
        if not identity:
            identity = ProvidedIdentity.create(
                db,
                data={
                    "privacy_request_id": None,
                    "field_name": ProvidedIdentityType.email.value,
                    "hashed_value": ProvidedIdentity.hash_value(identity_data.email),
                    "encrypted_value": {"value": identity_data.email},
                },
            )
    elif (
        target_identity_type == ProvidedIdentityType.phone_number.value
        and identity_data.phone_number
    ):
        identity = ProvidedIdentity.filter(
            db=db,
            conditions=(
                (ProvidedIdentity.field_name == ProvidedIdentityType.phone_number.value)
                & (
                    ProvidedIdentity.hashed_value.in_(
                        ProvidedIdentity.hash_value_for_search(
                            identity_data.phone_number
                        )
                    )
                )
                & (ProvidedIdentity.privacy_request_id.is_(None))
            ),
        ).first()
        if not identity:
            identity = ProvidedIdentity.create(
                db,
                data={
                    "privacy_request_id": None,
                    "field_name": ProvidedIdentityType.phone_number.value,
                    "hashed_value": ProvidedIdentity.hash_value(
                        identity_data.phone_number
                    ),
                    "encrypted_value": {"value": identity_data.phone_number},
                },
            )
    elif target_identity_type == ProvidedIdentityType.fides_user_device_id.value:
        identity = get_or_create_fides_user_device_id_provided_identity(
            db, identity_data
        )
    elif (
        target_identity_type == ProvidedIdentityType.external_id.value
        and identity_data.external_id
    ):
        identity = ProvidedIdentity.filter(
            db=db,
            conditions=(
                (ProvidedIdentity.field_name == ProvidedIdentityType.external_id.value)
                & (
                    ProvidedIdentity.hashed_value.in_(
                        ProvidedIdentity.hash_value_for_search(
                            identity_data.external_id
                        )
                    )
                )
                & (ProvidedIdentity.privacy_request_id.is_(None))
            ),
        ).first()
        if not identity:
            identity = ProvidedIdentity.create(
                db,
                data={
                    "privacy_request_id": None,
                    "field_name": ProvidedIdentityType.external_id.value,
                    "hashed_value": ProvidedIdentity.hash_value(
                        identity_data.external_id
                    ),
                    "encrypted_value": {"value": identity_data.external_id},
                },
            )

    else:
        raise HTTPException(
            HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Identity type not found in identity data",
        )
    return identity


def infer_target_identity_type(
    db: Session,
    identity_data: Identity,
) -> str:
    """
    Consent requests, unlike privacy requests, only accept 1 identity type: email,
    phone, external_id, or fides_user_device_id. These identity types are configurable
    as optional/required within the privacy center config.json. If both email and phone
    identity types are provided, we'll use the identity type defined in
    CONFIG.notifications.notification_service_type. Otherwise, the fallback order is
    email, phone_number, external_id, and finally fides_user_device_id.
    """
    target_identity_type = ""

    if identity_data.email and identity_data.phone_number:
        messaging_method = get_messaging_method(
            ConfigProxy(db).notifications.notification_service_type
        )
        if messaging_method == MessagingMethod.EMAIL:
            target_identity_type = ProvidedIdentityType.email.value
        elif messaging_method == MessagingMethod.SMS:
            target_identity_type = ProvidedIdentityType.phone_number.value
        else:
            target_identity_type = ProvidedIdentityType.email.value
    elif identity_data.email:
        target_identity_type = ProvidedIdentityType.email.value
    elif identity_data.phone_number:
        target_identity_type = ProvidedIdentityType.phone_number.value
    elif identity_data.external_id:
        target_identity_type = ProvidedIdentityType.external_id.value
    elif identity_data.fides_user_device_id:
        # If no other identity is provided, use the Fides User Device ID
        target_identity_type = ProvidedIdentityType.fides_user_device_id.value

    return target_identity_type


def _get_consent_request_and_provided_identity(
    db: Session,
    consent_request_id: str,
    verification_code: Optional[str],
) -> Tuple[ConsentRequest, ProvidedIdentity]:
    """Verifies the consent request and verification code, then return the ProvidedIdentity if successful."""
    consent_request = ConsentRequest.get_by_key_or_id(
        db=db, data={"id": consent_request_id}
    )

    if not consent_request:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="Consent request not found",
        )

    if not ConfigProxy(db).execution.disable_consent_identity_verification:
        try:
            consent_request.verify_identity(
                db,
                verification_code,
            )
        except IdentityVerificationException as exc:
            raise HTTPException(status_code=HTTP_400_BAD_REQUEST, detail=exc.message)
        except PermissionError as exc:
            logger.info(
                "Invalid verification code provided for {}.", consent_request.id
            )
            raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail=exc.args[0])

    provided_identity: ProvidedIdentity | None = ProvidedIdentity.get_by_key_or_id(
        db,
        data={"id": consent_request.provided_identity_id},
    )

    # It shouldn't be possible to hit this because the cascade delete of the identity
    # data would also delete the consent_request, but including this as a safety net.
    if not provided_identity:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="No identity found for consent request id",
        )

    return consent_request, provided_identity


def _prepare_consent_report(
    db: Session,
    consent: Consent,
) -> ConsentReport:
    """Enhances a consent request with identity, created and updated timestamps."""
    provided_identity = ProvidedIdentity.get_by(
        db=db,
        field="id",
        value=consent.provided_identity_id,
    )
    consent.identity = provided_identity.as_identity_schema()  # type: ignore[union-attr]
    report = ConsentReport.model_validate(consent)
    return report


def _prepare_consent_preferences(
    db: Session,
    provided_identity: ProvidedIdentity,
) -> ConsentPreferences:
    """Returns consent preferences for the identity given."""
    consent_records: List[Consent] = (
        Consent.filter(
            db=db, conditions=Consent.provided_identity_id == provided_identity.id
        )
        .order_by(Consent.updated_at)
        .all()
    )

    if not consent_records:
        return ConsentPreferences(consent=None)

    return ConsentPreferences(
        consent=[
            ConsentSchema(
                data_use=x.data_use,
                data_use_description=x.data_use_description,
                opt_in=x.opt_in,
                has_gpc_flag=x.has_gpc_flag,
                conflicts_with_gpc=x.conflicts_with_gpc,
            )
            for x in consent_records
        ],
    )



from typing import Any, Dict, List, Optional

import ahocorasick  # type: ignore
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.util.cache import FidesopsRedis, get_cache
from fides.config import get_config

AUTOMATON_SIGNAL_CACHE_KEY = "DECRYPTED_IDENTITY_AUTOMATON__CACHE_SIGNAL"
CONFIG = get_config()
_automaton = None


def get_decrypted_identities_automaton(
    db: Session, check_for_cache: bool = False
) -> ahocorasick.Automaton:  # pylint: disable=c-extension-no-member
    """
    Return a singleton Automaton that we can use for efficient fuzzy search.

    If check_for_cache is true, we refresh the automaton if automaton age is > 3 hrs.
    Automatons are only refreshed during new privacy requests (not during search) to reduce chance of slow performance
    during search.

    More background:

    Because we use SQLAlchemy-level AES/GCM encryption to write identity data to our ProvidedIdentity table,
    we cannot implement fuzzy search at the DB-level. No tools exist that support an equivalent AES/GCM decryption
    method within Postgres.

    Instead, we implement fuzzy search by decrypting identity data at the app-level. We do this by storing
    decrypted identity data in a singleton that refreshes every 3 hrs.

    We also manually write to the singleton when new privacy requests are created so that we do not miss newer values.
    """
    global _automaton  # pylint: disable=W0603
    if _automaton is None:
        logger.info(
            "Automaton does not yet exist. Proceeding to build automaton with decrypted identities..."
        )
        _automaton = build_automaton(db)
    elif check_for_cache and get_should_refresh_automaton():
        logger.info(
            "Automaton has expired. Proceeding to build new automaton with decrypted identities..."
        )
        _automaton = build_automaton(db)

    # Else use global pre-existing singleton
    return _automaton


def manually_reset_automaton() -> None:
    """Manually set our global _automaton singleton to None. Used for testing"""
    global _automaton  # pylint: disable=W0603
    _automaton = None


def build_automaton(
    db: Session,
) -> ahocorasick.Automaton:  # pylint: disable=c-extension-no-member
    """
    Builds automaton in this format: {"decrypted identity val", ["req_id_1", "req_id_2"]}
    """
    # Local import to avoid circular dependencies
    from fides.api.models.privacy_request import PrivacyRequest

    logger.debug("Creating new automaton...")
    automaton = ahocorasick.Automaton()  # pylint: disable=c-extension-no-member
    all_privacy_requests: List[PrivacyRequest] = db.query(PrivacyRequest).yield_per(1000)  # type: ignore
    for request in all_privacy_requests:
        _add_decrypted_identities_to_automaton(request.get_persisted_identity().__dict__, request.id, automaton)  # type: ignore
    set_automaton_cache_signal()
    return automaton


def add_identity_to_automaton(
    automaton: ahocorasick.Automaton,  # pylint: disable=c-extension-no-member
    request_id: str,
    identities: Optional[Dict[str, Any]],
) -> None:
    logger.info("Adding single identity to automaton...")
    _add_decrypted_identities_to_automaton(identities, request_id, automaton)


def set_automaton_cache_signal() -> None:
    """Set a signal we can check to determine whether we should refresh our decrypted identity automaton"""
    cache: FidesopsRedis = get_cache()
    logger.info("Setting automaton cache signal")
    cache.set_with_autoexpire(
        key=AUTOMATON_SIGNAL_CACHE_KEY,
        value="true",
        expire_time=10800,  # 3 hrs
    )


def remove_refresh_automaton_signal() -> None:
    """Remove should refresh automaton signal from cache for testing"""
    cache: FidesopsRedis = get_cache()
    cache.delete_keys_by_prefix(AUTOMATON_SIGNAL_CACHE_KEY)


def get_should_refresh_automaton() -> bool:
    """Returns whether we should refresh our decrypted identity automaton"""
    cache: FidesopsRedis = get_cache()
    result = cache.get(AUTOMATON_SIGNAL_CACHE_KEY)
    if result:
        return False
    return True


def _add_decrypted_identities_to_automaton(
    identities: Optional[Dict[str, Any]],
    request_id: str,
    automaton: ahocorasick.Automaton,  # pylint: disable=c-extension-no-member
) -> None:
    if not identities or not identities.items():
        return
    for key, value in identities.items():  # pylint: disable=W0612
        if value:
            if automaton.exists(value):
                existing: List[str] = automaton.get(value)  # value = decrypted identity
                existing.append(str(request_id))
                # overwrites the previously found key, updates with new request id
                automaton.add_word(value, existing)
            else:
                automaton.add_word(value, [request_id])



"""Module that adds interactions with aws"""

from functools import update_wrapper
from typing import Any, Callable, Dict, List, Optional

import boto3
from botocore.exceptions import ClientError
from fideslang.models import (
    Dataset,
    DatasetCollection,
    DatasetField,
    System,
    SystemMetadata,
)
from loguru import logger

from fides.connectors.models import (
    AWSConfig,
    ConnectorAuthFailureException,
    ConnectorFailureException,
)
from fides.core.utils import generate_unique_fides_key


def get_aws_client(service: str, aws_config: Optional[AWSConfig]) -> Any:
    """
    Creates boto3 client for a given service. A config is optional
    to allow for environment variable configuration.
    """
    config_dict = aws_config.model_dump(mode="json") if aws_config else {}
    service_client = boto3.client(
        service,
        **config_dict,
    )  # type: ignore
    return service_client


def handle_common_aws_errors(func: Callable) -> Callable:
    """
    Function decorator which handles common errors for aws calls.
    Classifies exceptions based on error codes returned by the client.
    """

    def wrapper_func(*args, **kwargs) -> Any:  # type: ignore
        try:
            return func(*args, **kwargs)
        except ClientError as error:
            logger.warning(error.response["Error"]["Code"])
            if error.response["Error"]["Code"] in [
                "InvalidClientTokenId",
                "SignatureDoesNotMatch",
            ]:
                raise ConnectorAuthFailureException(error.response["Error"]["Message"])
            if error.response["Error"]["Code"] in [
                "AccessDenied",
                "AccessDeniedException",
            ]:
                logger.warning(error.response["Error"]["Message"])
                return []
            raise ConnectorFailureException(error.response["Error"]["Message"])

    return update_wrapper(wrapper_func, func)


@handle_common_aws_errors
def validate_credentials(aws_config: Optional[AWSConfig]) -> None:
    """
    Calls the get_caller_identity sts api to validate aws credentials
    """
    client = get_aws_client("sts", aws_config)
    client.get_caller_identity()


@handle_common_aws_errors
def describe_redshift_clusters(client: Any) -> Dict[str, List[Dict]]:  # type: ignore
    """
    Returns describe_clusters response given a 'redshift' boto3 client.
    """
    describe_clusters = client.describe_clusters()
    return describe_clusters


@handle_common_aws_errors
def describe_rds_clusters(client: Any) -> Dict[str, List[Dict]]:  # type: ignore
    """
    Returns describe_db_clusters response given a rds boto3 client.
    """
    describe_clusters = client.describe_db_clusters()
    return describe_clusters


@handle_common_aws_errors
def describe_rds_instances(client: Any) -> Dict[str, List[Dict]]:  # type: ignore
    """
    Returns describe_db_instances response given a 'rds' boto3 client.
    """
    describe_instances = client.describe_db_instances()
    return describe_instances


@handle_common_aws_errors
def describe_dynamo_tables(client: Any, table_names: List[str]) -> List[Dict]:  # type: ignore
    """
    Returns describe_table response given a 'dynamodb' boto3 client.
    """
    describe_tables = []
    for table_name in table_names:
        described_table = describe_dynamo_table(client, table_name)
        if isinstance(described_table, dict):
            describe_tables.append(described_table)

    return describe_tables


@handle_common_aws_errors
def describe_dynamo_table(client: Any, table_name: str) -> Dict:  # type: ignore
    """
    Returns describe_table response given a 'dynamodb' boto3 client.
    """
    described_table = client.describe_table(TableName=table_name)
    described_table["Table"]["Fields"] = scan_dynamo_table(client, table_name)

    return described_table["Table"]


@handle_common_aws_errors
def get_dynamo_tables(client: Any) -> List[str]:  # type: ignore
    """
    Returns a list of table names response given a 'rds' boto3 client.
    """
    list_tables = client.list_tables()
    table_names = list_tables["TableNames"]
    next_page_exists = "LastEvaluatedTableName" in list_tables
    while next_page_exists:
        last_evaluated_table_name = list_tables["LastEvaluatedTableName"]
        list_tables = client.list_tables(
            ExclusiveStartTableName=last_evaluated_table_name
        )
        table_names.extend(list_tables["TableNames"])
        next_page_exists = "LastEvaluatedTableName" in list_tables
    return table_names


@handle_common_aws_errors
def scan_dynamo_table(client: Any, table_name: str, num_samples: int = 30) -> List[str]:  # type: ignore
    """
    Returns describe_table response given a 'dynamodb' boto3 client.
    """
    table_scan = client.scan(TableName=table_name, Limit=num_samples)
    fields = set()
    for item in table_scan["Items"]:
        for field, _ in item.items():
            fields.add(field)

    return list(fields)


@handle_common_aws_errors
def get_tagging_resources(client: Any) -> List[Dict]:  # type: ignore
    """
    Returns a list of resources given a 'resourcegroupstaggingapi' boto3 client.
    """
    paginator = client.get_paginator("get_resources")
    resources = [
        resource
        for page in paginator.paginate()
        for resource in page["ResourceTagMappingList"]
    ]
    return resources


def create_dynamodb_dataset(
    described_dynamo_tables: List[Dict],
    organization_key: str = "default_organization",
    single_dataset: bool = False,
) -> List[Dataset]:
    """
    Given "describe_table" response(s), build a dataset object to represent
    each dynamodb table, returning a fides dataset.
    """
    # TODO: add something for improved dataset uniqueness, i.e. region/account
    if single_dataset:
        dataset_name = "DynamoDB"
        unique_dataset_name = generate_unique_fides_key(dataset_name, "", "")
        datasets = [
            Dataset(
                name=dataset_name,
                fides_key=unique_dataset_name,
                organization_fides_key=organization_key,
                collections=[
                    DatasetCollection(
                        name=collection["TableName"],
                        # description=collection["TableArn"],
                        fields=[
                            DatasetField(
                                name=field,
                                description=f"Fides Generated Description for Column: {field}",
                                data_categories=[],
                                # TODO: include a fieldsmeta if the field is a primary key or secondary sort (and test for both)
                            )
                            for field in collection["Fields"]
                        ],
                    )
                    for collection in described_dynamo_tables
                ],
            )
        ]
    else:
        datasets = [
            Dataset(
                name=collection["TableName"],
                fides_key=generate_unique_fides_key(collection["TableName"], "", ""),
                organization_fides_key=organization_key,
                collections=[
                    DatasetCollection(
                        name=collection["TableName"],
                        # description=collection["TableArn"],
                        fields=[
                            DatasetField(
                                name=field,
                                description=f"Fides Generated Description for Column: {field}",
                                data_categories=[],
                                # TODO: include a fieldsmeta if the field is a primary key or secondary sort (and test for both)
                            )
                            for field in collection["Fields"]
                        ],
                    )
                ],
            )
            for collection in described_dynamo_tables
        ]
    return datasets


def create_redshift_systems(
    describe_clusters: Dict[str, List[Dict]], organization_key: str
) -> List[System]:
    """
    Given a "describe_clusters" response, build a system object which represents
    each cluster.
    """
    redshift_systems = [
        System(
            fides_key=cluster["ClusterIdentifier"],
            name=cluster["ClusterIdentifier"],
            description=f"Fides Generated Description for Redshift Cluster: {cluster['ClusterIdentifier']}",
            meta={pair["Key"]: pair["Value"] for pair in cluster.get("Tags", {})},
            system_type="redshift_cluster",
            organization_fides_key=organization_key,
            fidesctl_meta=SystemMetadata(
                endpoint_address=(
                    cluster["Endpoint"]["Address"] if cluster.get("Endpoint") else None
                ),
                endpoint_port=(
                    cluster["Endpoint"]["Port"] if cluster.get("Endpoint") else None
                ),
                resource_id=cluster["ClusterNamespaceArn"],
            ),
            privacy_declarations=[],
        )
        for cluster in describe_clusters["Clusters"]
    ]
    return redshift_systems


def create_rds_systems(
    describe_clusters: Dict[str, List[Dict]],
    describe_instances: Dict[str, List[Dict]],
    organization_key: str,
) -> List[System]:
    """
    Given "describe_clusters" and "describe_instances" responses, build a system object
    which represents each cluster or instance.

    A system is created for each cluster, but for instances we only create a system if
    it is not part of a cluster
    """
    rds_cluster_systems = [
        System(
            fides_key=cluster["DBClusterIdentifier"],
            name=cluster["DBClusterIdentifier"],
            description=f"Fides Generated Description for RDS Cluster: {cluster['DBClusterIdentifier']}",
            meta={pair["Key"]: pair["Value"] for pair in cluster.get("TagList", {})},
            system_type="rds_cluster",
            organization_fides_key=organization_key,
            fidesctl_meta=SystemMetadata(
                endpoint_address=cluster["Endpoint"],
                endpoint_port=cluster["Port"],
                resource_id=cluster["DBClusterArn"],
            ),
            privacy_declarations=[],
        )
        for cluster in describe_clusters["DBClusters"]
    ]
    rds_instances_systems = [
        System(
            fides_key=instance["DBInstanceIdentifier"],
            name=instance["DBInstanceIdentifier"],
            description=f"Fides Generated Description for RDS Instance: {instance['DBInstanceIdentifier']}",
            meta={pair["Key"]: pair["Value"] for pair in instance.get("TagList", {})},
            system_type="rds_instance",
            organization_fides_key=organization_key,
            fidesctl_meta=SystemMetadata(
                endpoint_address=(
                    instance["Endpoint"]["Address"]
                    if instance.get("Endpoint")
                    else None
                ),
                endpoint_port=(
                    instance["Endpoint"]["Port"] if instance.get("Endpoint") else None
                ),
                resource_id=instance["DBInstanceArn"],
            ),
            privacy_declarations=[],
        )
        for instance in describe_instances["DBInstances"]
        if not instance.get("DBClusterIdentifier")
    ]
    return rds_cluster_systems + rds_instances_systems


def create_resource_tagging_systems(
    resources: List[Dict],
    organization_key: str,
) -> List[System]:
    """
    Given a list of resources, build a list of systems object which represents
    each resource.
    """
    resource_generators = {
        "dynamodb": create_tagging_dynamodb_system,
        "s3": create_tagging_s3_system,
    }
    systems = []
    for resource in resources:
        arn_split = resource["ResourceARN"].split(":")
        arn_resource_type = arn_split[2]
        resource_generator = resource_generators.get(arn_resource_type)
        if resource_generator:
            generated_system = resource_generator(resource, organization_key)
            if generated_system:
                systems.append(generated_system)
    return systems


def create_tagging_dynamodb_system(
    resource: Dict[str, Any], organization_key: str
) -> Optional[System]:
    """
    Given an AWS dynamodb resource, returns a System representation
    for dynamodb tables.
    """
    arn = resource["ResourceARN"]
    arn_split = arn.split(":")
    resource_name = arn_split[5]

    if resource_name.startswith("table/"):
        table_name = resource_name[len("table/") :]
        system = System(
            fides_key=table_name,
            name=table_name,
            description=f"Fides Generated Description for DynamoDb table: {table_name}",
            meta={pair["Key"]: pair["Value"] for pair in resource.get("Tags", {})},
            system_type="dynamodb_table",
            organization_fides_key=organization_key,
            fidesctl_meta=SystemMetadata(
                resource_id=arn,
            ),
            privacy_declarations=[],
        )
        return system
    return None


def create_tagging_s3_system(
    resource: Dict[str, Any], organization_key: str
) -> Optional[System]:
    """
    Given an AWS s3 resource, returns a System representation
    for s3 buckets.
    """
    arn = resource["ResourceARN"]
    arn_split = arn.split(":")
    resource_name = arn_split[5]

    bucket_name = resource_name.split("/")[0]
    system = System(
        fides_key=bucket_name,
        name=bucket_name,
        description=f"Fides Generated Description for S3 bucket: {bucket_name}",
        meta={pair["Key"]: pair["Value"] for pair in resource.get("Tags", {})},
        system_type="s3_bucket",
        organization_fides_key=organization_key,
        fidesctl_meta=SystemMetadata(
            resource_id=arn,
        ),
        privacy_declarations=[],
    )
    return system



"""Adds deleted_at and deleted_by to PrivacyRequest

Revision ID: 75bb9ee843f5
Revises: 68c590ff6e89
Create Date: 2024-09-19 14:35:30.510909

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "75bb9ee843f5"
down_revision = "68c590ff6e89"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacyrequest",
        sa.Column("deleted_at", sa.DateTime(timezone=True), nullable=True),
    )
    op.add_column("privacyrequest", sa.Column("deleted_by", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyrequest", "deleted_by")
    op.drop_column("privacyrequest", "deleted_at")
    # ### end Alembic commands ###



"""remove delivery mechanism

Revision ID: deb97d9393f3
Revises: ed46521679fb
Create Date: 2023-05-25 16:47:16.566034

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "deb97d9393f3"
down_revision = "ed46521679fb"
branch_labels = None
depends_on = None


def upgrade():
    """Remove required delivery mechanism field and rename/add a lot of new fields related to experience configs"""
    op.drop_index(
        "ix_privacyexperienceconfig_delivery_mechanism",
        table_name="privacyexperienceconfig",
    )
    op.drop_index(
        "ix_privacyexperienceconfighistory_delivery_mechanism",
        table_name="privacyexperienceconfighistory",
    )

    op.drop_column("privacyexperienceconfig", "delivery_mechanism")
    op.drop_column("privacyexperienceconfighistory", "delivery_mechanism")

    op.drop_column("privacyexperience", "delivery_mechanism")
    op.drop_column("privacyexperiencehistory", "delivery_mechanism")

    op.drop_column("privacyexperienceconfig", "acknowledgement_button_label")
    op.drop_column("privacyexperienceconfig", "banner_description")
    op.drop_column("privacyexperienceconfig", "banner_title")
    op.drop_column("privacyexperienceconfig", "component_title")
    op.drop_column("privacyexperienceconfig", "component_description")
    op.drop_column("privacyexperienceconfig", "confirmation_button_label")
    op.drop_column("privacyexperienceconfig", "link_label")

    op.add_column(
        "privacyexperienceconfig",
        sa.Column("accept_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("acknowledge_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_enabled", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig", sa.Column("description", sa.String(), nullable=True)
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("privacy_preferences_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("privacy_policy_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("privacy_policy_url", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("save_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig", sa.Column("title", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_privacyexperienceconfig_banner_enabled"),
        "privacyexperienceconfig",
        ["banner_enabled"],
        unique=False,
    )

    op.drop_column("privacyexperienceconfighistory", "acknowledgement_button_label")
    op.drop_column("privacyexperienceconfighistory", "banner_description")
    op.drop_column("privacyexperienceconfighistory", "banner_title")
    op.drop_column("privacyexperienceconfighistory", "component_description")
    op.drop_column("privacyexperienceconfighistory", "confirmation_button_label")
    op.drop_column("privacyexperienceconfighistory", "component_title")
    op.drop_column("privacyexperienceconfighistory", "link_label")

    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("accept_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("acknowledge_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("banner_enabled", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("description", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("privacy_preferences_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("privacy_policy_link_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("privacy_policy_url", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("save_button_label", sa.String(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory", sa.Column("title", sa.String(), nullable=True)
    )
    op.create_index(
        op.f("ix_privacyexperienceconfighistory_banner_enabled"),
        "privacyexperienceconfighistory",
        ["banner_enabled"],
        unique=False,
    )


def downgrade():
    op.add_column(
        "privacyexperiencehistory",
        sa.Column(
            "delivery_mechanism", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("component_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "acknowledgement_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "confirmation_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "component_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("link_label", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "delivery_mechanism", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfighistory_banner_enabled"),
        table_name="privacyexperienceconfighistory",
    )
    op.create_index(
        "ix_privacyexperienceconfighistory_delivery_mechanism",
        "privacyexperienceconfighistory",
        ["delivery_mechanism"],
        unique=False,
    )

    op.drop_column("privacyexperienceconfighistory", "title")
    op.drop_column("privacyexperienceconfighistory", "save_button_label")
    op.drop_column("privacyexperienceconfighistory", "privacy_policy_url")
    op.drop_column("privacyexperienceconfighistory", "privacy_policy_link_label")
    op.drop_column("privacyexperienceconfighistory", "privacy_preferences_link_label")
    op.drop_column("privacyexperienceconfighistory", "description")
    op.drop_column("privacyexperienceconfighistory", "banner_enabled")
    op.drop_column("privacyexperienceconfighistory", "acknowledge_button_label")
    op.drop_column("privacyexperienceconfighistory", "accept_button_label")

    op.add_column(
        "privacyexperienceconfig",
        sa.Column("component_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "acknowledgement_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "confirmation_button_label",
            sa.VARCHAR(),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "component_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("link_label", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "delivery_mechanism", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("banner_title", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "banner_description", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )
    op.drop_index(
        op.f("ix_privacyexperienceconfig_banner_enabled"),
        table_name="privacyexperienceconfig",
    )
    op.create_index(
        "ix_privacyexperienceconfig_delivery_mechanism",
        "privacyexperienceconfig",
        ["delivery_mechanism"],
        unique=False,
    )

    op.drop_column("privacyexperienceconfig", "title")
    op.drop_column("privacyexperienceconfig", "save_button_label")
    op.drop_column("privacyexperienceconfig", "privacy_policy_url")
    op.drop_column("privacyexperienceconfig", "privacy_policy_link_label")
    op.drop_column("privacyexperienceconfig", "privacy_preferences_link_label")
    op.drop_column("privacyexperienceconfig", "description")
    op.drop_column("privacyexperienceconfig", "banner_enabled")
    op.drop_column("privacyexperienceconfig", "acknowledge_button_label")
    op.drop_column("privacyexperienceconfig", "accept_button_label")
    op.add_column(
        "privacyexperience",
        sa.Column(
            "delivery_mechanism", sa.VARCHAR(), autoincrement=False, nullable=True
        ),
    )



"""
Contains all of the generic CRUD endpoints that can be
generated programmatically for each resource.
"""

from collections import defaultdict
from typing import Any, Dict, List, Tuple, Type, TypeVar

from fastapi import HTTPException
from loguru import logger as log
from sqlalchemy import and_, column
from sqlalchemy import delete as _delete
from sqlalchemy import or_
from sqlalchemy import update as _update
from sqlalchemy.dialects.postgresql import Insert as _insert
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.sql import Select
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

from fides.api.models.sql_models import (  # type: ignore[attr-defined]
    CustomField,
    CustomFieldDefinition,
    FidesBase,
    ResourceTypes,
)
from fides.api.util import errors

# Helps return type be linked to the type of the parameter
T = TypeVar("T", bound="FidesBase")


# CRUD Functions
async def create_resource(
    sql_model: Type[T], resource_dict: Dict, async_session: AsyncSession
) -> T:
    """Create a resource in the database."""
    with log.contextualize(
        sql_model=sql_model.__name__, fides_key=resource_dict["fides_key"]
    ):

        existing_resource = await get_resource(
            sql_model, resource_dict["fides_key"], async_session, raise_not_found=False
        )

        if existing_resource is not None:
            already_exists_error = errors.AlreadyExistsError(
                sql_model.__name__, resource_dict["fides_key"]
            )
            log.bind(error=already_exists_error.detail["error"]).info(  # type: ignore[index]
                "Failed to insert resource"
            )
            raise already_exists_error

        async with async_session.begin():
            try:
                log.debug("Creating resource")
                query = _insert(sql_model.__table__).values(resource_dict)
                await async_session.execute(query)
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).info(  # type: ignore[index]
                    "Failed to create resource"
                )
                raise sa_error

        return await get_resource(sql_model, resource_dict["fides_key"], async_session)


async def get_custom_fields_filtered(
    async_session: AsyncSession,
    resource_types_to_ids: Dict[ResourceTypes, List[str]] = defaultdict(list),
) -> FidesBase:
    """
    Utility function to construct a filtered query for custom field values based on provided mapping of
    resource types to resource IDs.

    Only custom fields with an "active" CustomFieldDefinition are returned.

    This is for use in bulk querying of custom fields, to avoid multiple round trips to the db.
    """
    with log.contextualize(model=CustomField):
        async with async_session.begin():
            try:
                log.debug("Fetching resource")
                query = select(
                    CustomField.resource_id,
                    CustomField.value,
                    CustomFieldDefinition.resource_type,
                    CustomFieldDefinition.name,
                    CustomFieldDefinition.field_type,
                ).join(
                    CustomFieldDefinition,
                    CustomFieldDefinition.id == CustomField.custom_field_definition_id,
                )

                criteria = [
                    and_(
                        CustomFieldDefinition.resource_type == resource_type.value,
                        CustomField.resource_id.in_(resource_ids),
                        # pylint: disable=singleton-comparison
                        CustomFieldDefinition.active == True,
                    )
                    for resource_type, resource_ids in resource_types_to_ids.items()
                ]
                query = query.where(or_(False, *criteria))
                result = await async_session.execute(query)
                return result.mappings().all()
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).error(  # type: ignore[index]
                    "Failed to fetch custom fields"
                )
                raise sa_error


async def get_resource(
    sql_model: Type[T],
    fides_key: str,
    async_session: AsyncSession,
    raise_not_found: bool = True,
) -> T:
    """
    Get a resource from the database by its FidesKey.

    Returns a SQLAlchemy model of that resource.
    """
    with log.contextualize(sql_model=sql_model.__name__, fides_key=fides_key):
        async with async_session.begin():
            try:
                log.debug("Fetching resource")
                query = select(sql_model).where(sql_model.fides_key == fides_key)
                result = await async_session.execute(query)
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).info(  # type: ignore[index]
                    "Failed to fetch resource"
                )
                raise sa_error

            sql_resource = result.scalars().first()
            if sql_resource is None and raise_not_found:
                not_found_error = errors.NotFoundError(sql_model.__name__, fides_key)
                log.bind(error=not_found_error.detail["error"]).info("Resource not found")  # type: ignore[index]
                raise not_found_error

            return sql_resource


async def get_resource_with_custom_fields(
    sql_model: Type[T], fides_key: str, async_session: AsyncSession
) -> Dict[str, Any]:
    """Get a resource from the databse by its FidesKey including it's custom fields.

    Returns a dictionary of that resource.
    """
    resource: T = await get_resource(sql_model, fides_key, async_session)
    resource_dict = resource.__dict__
    resource_dict.pop("_sa_instance_state", None)

    with log.contextualize(sql_model=sql_model.__name__, fides_key=fides_key):
        async with async_session.begin():
            try:
                log.debug("Fetching custom fields for resource")
                query = (
                    select(CustomFieldDefinition.name, CustomField.value)
                    .join(
                        CustomField,
                        CustomField.custom_field_definition_id
                        == CustomFieldDefinition.id,
                    )
                    .where(
                        (CustomField.resource_id == resource.fides_key)
                        & (  # pylint: disable=singleton-comparison
                            CustomFieldDefinition.active == True
                        )
                    )
                )
                result = await async_session.execute(query)
            except SQLAlchemyError:
                sa_error = errors.QueryError()
                log.bind(error=sa_error.detail["error"]).info(  # type: ignore[index]
                    "Failed to fetch custom fields"
                )
                raise sa_error

            custom_fields = result.mappings().all()

    if not custom_fields:
        return resource_dict

    for field in custom_fields:
        if field["name"] in resource_dict:
            resource_dict[field["name"]] = (
                f"{resource_dict[field['name']]}, {', '.join(field['value'])}"
            )
        else:
            resource_dict[field["name"]] = ", ".join(field["value"])

    return resource_dict


async def list_resource(sql_model: Type[T], async_session: AsyncSession) -> List[T]:
    """
    Get a list of all of the resources of this type from the database.

    Returns a list of SQLAlchemy models of that resource type.
    """
    query = select(sql_model)
    return await list_resource_query(async_session, query, sql_model)


async def list_resource_query(
    async_session: AsyncSession, query: Select, sql_model: Type[T]
) -> List[T]:
    """
    Utility function to wrap a select query in generic "list_resource" execution handling.
    Wrapping includes execution against the DB session, logging and error handling.
    """

    with log.contextualize(sql_model=sql_model.__name__):
        async with async_session.begin():
            try:
                log.debug("Fetching resources")
                result = await async_session.execute(query)
                sql_resources = result.scalars().all()
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).info(  # type: ignore[index]
                    "Failed to fetch resources"
                )
                raise error

            return sql_resources


async def update_resource(
    sql_model: Type[T], resource_dict: Dict, async_session: AsyncSession
) -> Dict:
    """Update a resource in the database by its fides_key."""

    with log.contextualize(
        sql_model=sql_model.__name__, fides_key=resource_dict["fides_key"]
    ):
        await get_resource(sql_model, resource_dict["fides_key"], async_session)

        async with async_session.begin():
            try:
                log.debug("Updating resource")
                await async_session.execute(
                    _update(sql_model.__table__)
                    .where(sql_model.fides_key == resource_dict["fides_key"])
                    .values(resource_dict)
                )
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).info(  # type: ignore[index]
                    "Failed to update resource"
                )
                raise error

        return await get_resource(sql_model, resource_dict["fides_key"], async_session)


async def upsert_resources(
    sql_model: Type[T], resource_dicts: List[Dict], async_session: AsyncSession
) -> Tuple[int, int]:
    """
    Insert new resources into the database. If a resource already exists,
    update it by it's fides_key.

    Returns a Tuple containing the counts of inserted and updated rows.
    """

    with log.contextualize(
        sql_model=sql_model.__name__,
        fides_keys=[resource["fides_key"] for resource in resource_dicts],
    ):
        async with async_session.begin():
            try:
                log.debug("Upserting resources")
                insert_stmt = (
                    _insert(sql_model.__table__)
                    .values(resource_dicts)
                    .returning(
                        (column("xmax") == 0),  # Row was inserted
                        (column("xmax") != 0),  # Row was updated
                    )
                )

                excluded = dict(insert_stmt.excluded.items())  # type: ignore[attr-defined]
                excluded.pop("id", None)  # If updating, don't update the "id"

                result = await async_session.execute(
                    insert_stmt.on_conflict_do_update(
                        index_elements=["fides_key"],
                        set_=excluded,
                    )
                )

                inserts, updates = 0, 0
                for xmax in result:
                    inserts += 1 if xmax[0] else 0
                    updates += 1 if xmax[1] else 0

                return (inserts, updates)

            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).info(  # type: ignore[index]
                    "Failed to upsert resources"
                )
                log.exception(error)
                raise error


async def delete_resource(
    sql_model: Type[T], fides_key: str, async_session: AsyncSession
) -> T:
    """
    Delete a resource by its fides_key.

    If the resource has child keys referring to it, also delete those.
    """

    with log.contextualize(sql_model=sql_model.__name__, fides_key=fides_key):
        sql_resource = await get_resource(sql_model, fides_key, async_session)

        async with async_session.begin():
            try:
                # Automatically delete related resources if they are CTL objects
                if hasattr(sql_model, "parent_key"):
                    log.debug("Deleting resource and its children")
                    query = (
                        _delete(sql_model.__table__)
                        .where(
                            or_(
                                sql_model.fides_key == fides_key,
                                sql_model.fides_key.like(f"{fides_key}.%"),
                            )
                        )
                        .execution_options(synchronize_session=False)
                    )
                else:
                    log.debug("Deleting resource")
                    query = _delete(sql_model.__table__).where(
                        sql_model.fides_key == fides_key
                    )
                await async_session.execute(query)
            except IntegrityError as err:
                raw_error_text: str = err.orig.args[0]

                if "violates foreign key constraint" in raw_error_text:
                    error_message = "Failed to delete resource! Foreign key constraint found, try deleting related resources first."
                else:
                    error_message = "Failed to delete resource!"

                log.bind(error="SQL Query integrity error!").error(raw_error_text)
                raise HTTPException(
                    status_code=HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=error_message,
                )
            except SQLAlchemyError:
                error = errors.QueryError()
                log.bind(error=error.detail["error"]).info(  # type: ignore[index]
                    "Failed to delete resource"
                )
                raise error

        return sql_resource



from __future__ import annotations

from typing import Any, Dict, List, Optional

import httpx
from httpx import AsyncClient, Client, HTTPStatusError, Request, RequestError, Timeout
from loguru import logger

from fides.api.schemas.privacy_request import (
    PrivacyRequestCreate,
    PrivacyRequestResponse,
    PrivacyRequestStatus,
)
from fides.api.schemas.redis_cache import Identity
from fides.api.schemas.user import UserLogin
from fides.api.service.privacy_request.request_service import poll_server_for_completion
from fides.api.util.collection_util import Row
from fides.api.util.errors import FidesError
from fides.api.util.wrappers import sync
from fides.common.api.v1 import urn_registry as urls

COMPLETION_STATUSES = [
    PrivacyRequestStatus.complete,
    PrivacyRequestStatus.canceled,
    PrivacyRequestStatus.error,
    PrivacyRequestStatus.denied,
]


class FidesClient:
    """
    A helper client to broker communications between Fides servers.
    """

    def __init__(
        self,
        uri: str,
        username: str,
        password: str,
        connection_read_timeout: float = 30.0,
    ):
        # Enable setting a custom `read` timeout
        # to account for privacy request executions
        self.session = Client(timeout=Timeout(5.0, read=connection_read_timeout))

        self.uri = uri
        self.username = username
        self.password = password
        self.token = None

    def login(self) -> None:
        ul: UserLogin = UserLogin(username=self.username, password=self.password)
        logger.info(
            "Logging in to remote fides {} with username '{}'...",
            self.uri,
            self.username,
        )
        try:
            response = httpx.post(
                f"{self.uri}{urls.V1_URL_PREFIX}{urls.LOGIN}",
                json=ul.model_dump(mode="json"),
            )
        except RequestError as e:
            logger.error("Error logging in on remote Fides {}: {}", self.uri, str(e))
            raise e

        if response.is_success:
            self.token = response.json()["token_data"]["access_token"]
            logger.info(
                "Successfully logged in to remote fides {} with username '{}'",
                self.uri,
                self.username,
            )
        else:
            logger.error("Error logging in on remote Fides {}", self.uri)
            response.raise_for_status()

    def authenticated_request(
        self,
        method: str,
        path: str,
        headers: Optional[Dict[str, Any]] = {},
        query_params: Optional[Dict[str, Any]] = {},
        data: Optional[Any] = None,
        json: Optional[Any] = None,
    ) -> Request:
        if not self.token:
            raise FidesError(
                f"Unable to create authenticated request. No token for Fides connector for server {self.uri}"
            )

        req: Request = self.session.build_request(
            method=method,
            url=f"{self.uri}{path}",
            headers=headers,
            params=query_params,
            data=data,
            json=json,
        )
        req.headers["Authorization"] = f"Bearer {self.token}"
        return req

    def create_privacy_request(
        self, external_id: Optional[str], identity: Identity, policy_key: str
    ) -> str:
        """
        Create privacy request on remote fides by hitting privacy request endpoint
        Returns the created privacy request ID
        """
        pr: PrivacyRequestCreate = PrivacyRequestCreate(
            external_id=external_id,
            identity=identity,
            policy_key=policy_key,
        )

        logger.info(
            "Creating privacy request with external_id {} on remote fides {}...",
            external_id,
            self.uri,
        )
        request: Request = self.authenticated_request(
            method="POST",
            path=urls.V1_URL_PREFIX + urls.PRIVACY_REQUEST_AUTHENTICATED,
            json=[pr.model_dump(mode="json")],
        )
        response = self.session.send(request)

        if not response.is_success:
            logger.error("Error creating privacy request on remote Fides {}", self.uri)
            response.raise_for_status()

        if response.json()["failed"]:
            # TODO better handle errored state here?
            raise FidesError(
                f"Failed privacy request creation on remote Fides {self.uri} with failure message: {response.json()['failed'][0]['message']}"
            )

        pr_id = response.json()["succeeded"][0]["id"]
        logger.info(
            "Successfully created privacy request with id {} and external_id {} on remote fides {}",
            pr_id,
            external_id,
            self.uri,
        )
        return pr_id

    @sync
    async def poll_for_request_completion(
        self,
        privacy_request_id: str,
        timeout: int,
        interval: int,
        async_client: AsyncClient | None = None,
    ) -> PrivacyRequestResponse:
        """
        Poll remote fides for status of privacy request with the given ID until it is complete.
        This is effectively a blocking call, i.e. it will block the current thread until
        it determines completion, or until timeout is reached.

        Returns the privacy request record, or error
        """

        if not self.token:
            raise FidesError(
                f"Unable to poll for request completion. No token for Fides connector for server {self.uri}"
            )

        logger.info(
            "Polling remote fides {} for completion of privacy request with id {}...",
            self.uri,
            privacy_request_id,
        )
        status: PrivacyRequestResponse = await poll_server_for_completion(
            privacy_request_id=privacy_request_id,
            server_url=self.uri,
            token=self.token,
            poll_interval_seconds=interval,
            timeout_seconds=timeout,
            client=async_client,
        )
        if status.status == PrivacyRequestStatus.error:
            raise FidesError(
                f"Privacy request [{privacy_request_id}] on remote Fides {self.uri} encountered an error. Look at the remote Fides for more information."
            )
        if status.status == PrivacyRequestStatus.canceled:
            raise FidesError(
                f"Privacy request [{privacy_request_id}] on remote Fides {self.uri} was canceled. Look at the remote Fides for more information."
            )
        if status.status == PrivacyRequestStatus.denied:
            raise FidesError(
                f"Privacy request [{privacy_request_id}] on remote Fides {self.uri} was denied. Look at the remote Fides for more information."
            )
        if status.status == PrivacyRequestStatus.complete:
            logger.info(
                "Privacy request [{}] is complete on remote Fides {}!",
                privacy_request_id,
                self.uri,
            )
            return status

        raise FidesError(
            f"Privacy request [{privacy_request_id}] on remote Fides {self.uri} is in an unknown state. Look at the remote Fides for more information."
        )

    def request_status(self, privacy_request_id: str = "") -> List[Dict[str, Any]]:
        """
        Return privacy request object that tracks its status
        """
        if privacy_request_id:
            logger.info(
                "Retrieving request status for privacy request {} on remote fides {}...",
                privacy_request_id,
                self.uri,
            )
        else:
            logger.info(
                "Retrieving request status for all privacy requests on remote fides {}...",
                self.uri,
            )

        request: Request = self.authenticated_request(
            method="GET",
            path=urls.V1_URL_PREFIX + urls.PRIVACY_REQUESTS,
            query_params=(
                {"request_id": privacy_request_id} if privacy_request_id else None
            ),
        )
        response = self.session.send(request)

        if not response.is_success:
            logger.error(
                "Error retrieving status of privacy request [{}] on remote Fides {}",
                privacy_request_id,
                self.uri,
            )
            response.raise_for_status()

        if privacy_request_id:
            logger.info(
                "Retrieved request status for privacy request {} on remote fides {}",
                privacy_request_id,
                self.uri,
            )
        else:
            logger.info(
                "Retrieved request status for all privacy requests on remote fides {}",
                self.uri,
            )
        return response.json()["items"]

    def retrieve_request_results(
        self, privacy_request_id: str, rule_key: str
    ) -> Dict[str, List[Row]]:
        """
        Retrieve the filtered access results on the remote fides associated with
        the given `privacy_request_id` and `rule_key`, by invoking the
        `privacy_request_data_transfer` endpoint on the remote Fides.

        Returns the filtered access results as a `Dict[str, List[Row]]
        """
        try:
            logger.info(
                "Retrieving request results for privacy request {} on remote fides {}...",
                privacy_request_id,
                self.uri,
            )
            request = self.authenticated_request(
                method="get",
                path=f"{urls.V1_URL_PREFIX}{urls.PRIVACY_REQUEST_TRANSFER_TO_PARENT.format(privacy_request_id=privacy_request_id, rule_key=rule_key)}",
                headers={"Authorization": f"Bearer {self.token}"},
            )
            response = self.session.send(request)
        except HTTPStatusError as e:
            logger.error(
                "Error retrieving data from child server for privacy request {}: {}",
                privacy_request_id,
                e,
            )

        if response.status_code != 200:
            logger.error(
                "Error retrieving data from child server for privacy request {}: {}",
                privacy_request_id,
                response.text,
            )
            return {}

        logger.info(
            "Retrieved request results for privacy request {} on remote fides {}",
            privacy_request_id,
            self.uri,
        )
        return response.json()



from typing import List, Optional

from pydantic import BaseModel


class FilterParams(BaseModel):
    """
    Generic parameters for filtering queries.
    """

    search: Optional[str] = None
    data_uses: Optional[List[str]] = None
    data_categories: Optional[List[str]] = None
    data_subjects: Optional[List[str]] = None



from typing import Dict, Optional

from fastapi import Form
from fastapi.openapi.models import OAuthFlows
from fastapi.security import OAuth2
from fastapi.security.utils import get_authorization_scheme_param
from pydantic import BaseModel
from starlette.requests import Request

from fides.api.common_exceptions import InvalidAuthorizationSchemeError


class AccessToken(BaseModel):
    """A wrapper for the access_code returned upon successful authentication"""

    access_token: str


# NOTE: Adapted from
#   https://github.com/tiangolo/fastapi/blob/master/fastapi/security/oauth2.py#L140
class OAuth2ClientCredentialsBearer(OAuth2):
    """Requires a valid OAuth2 bearer token using the client credentials flow, e.g.
    "Authorization: Bearer <token>".

    See /oauth/token for details on how to authenticate and receive a valid token.
    """

    def __init__(
        self,
        tokenUrl: str,
        scheme_name: Optional[str] = None,
        scopes: Optional[Dict[str, str]] = None,
        description: Optional[str] = None,
        auto_error: bool = True,
    ):
        if not scopes:
            scopes = {}
        flows = OAuthFlows(clientCredentials={"tokenUrl": tokenUrl, "scopes": scopes})
        super().__init__(
            flows=flows,
            scheme_name=scheme_name,
            description=description,
            auto_error=auto_error,
        )

    async def __call__(self, request: Request) -> Optional[str]:
        authorization: Optional[str] = request.headers.get("Authorization")
        if not authorization:
            raise InvalidAuthorizationSchemeError()
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            if self.auto_error:
                raise InvalidAuthorizationSchemeError()
            return None  # pragma: no cover
        return param


# NOTE: Adapted from
#   https://github.com/tiangolo/fastapi/blob/master/fastapi/security/oauth2.py#L13
# NOTE: This uses application/x-www-form-urlencoded (form encoding) instead of
#   JSON to follow the
# OAuth2 spec: https://datatracker.ietf.org/doc/html/rfc6749#section-4.4.2
class OAuth2ClientCredentialsRequestForm:
    """Request model used to authenticate via OAuth2 client credentials."""

    def __init__(
        self,
        grant_type: str = Form(None, regex="client_credentials"),
        scope: str = Form(""),
        client_id: Optional[str] = Form(None),
        client_secret: Optional[str] = Form(None),
    ):
        self.grant_type = grant_type
        self.scopes = scope.split()
        self.client_id = client_id
        self.client_secret = client_secret



from __future__ import annotations

import json
import secrets
import zipfile
from io import BytesIO
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import pandas as pd
from botocore.exceptions import ClientError, ParamValidationError
from fideslang.validation import AnyHttpUrlString
from loguru import logger

from fides.api.cryptography.cryptographic_util import bytes_to_b64_str
from fides.api.schemas.storage.storage import ResponseFormat, StorageSecrets
from fides.api.service.privacy_request.dsr_package.dsr_report_builder import (
    DsrReportBuilder,
)
from fides.api.service.storage.gcs import get_gcs_client
from fides.api.service.storage.s3 import (
    create_presigned_url_for_s3,
    generic_upload_to_s3,
)
from fides.api.service.storage.util import (
    LOCAL_FIDES_UPLOAD_DIRECTORY,
    get_local_filename,
)
from fides.api.util.aws_util import get_s3_client
from fides.api.util.cache import get_cache, get_encryption_cache_key
from fides.api.util.encryption.aes_gcm_encryption_scheme import (
    encrypt_to_bytes_verify_secrets_length,
)
from fides.api.util.storage_util import storage_json_encoder
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request import PrivacyRequest


def encrypt_access_request_results(data: Union[str, bytes], request_id: str) -> str:
    """Encrypt data with encryption key if provided, otherwise return unencrypted data"""
    cache = get_cache()
    encryption_cache_key = get_encryption_cache_key(
        privacy_request_id=request_id,
        encryption_attr="key",
    )
    if isinstance(data, bytes):
        data = data.decode(CONFIG.security.encoding)

    encryption_key: str | None = cache.get(encryption_cache_key)
    if not encryption_key:
        return data

    bytes_encryption_key: bytes = encryption_key.encode(
        encoding=CONFIG.security.encoding
    )
    nonce: bytes = secrets.token_bytes(CONFIG.security.aes_gcm_nonce_length)
    # b64encode the entire nonce and the encrypted message together
    return bytes_to_b64_str(
        nonce
        + encrypt_to_bytes_verify_secrets_length(data, bytes_encryption_key, nonce)
    )


def write_to_in_memory_buffer(
    resp_format: str, data: Dict[str, Any], privacy_request: PrivacyRequest
) -> BytesIO:
    """Write JSON/CSV data to in-memory file-like object to be passed to S3 or GCS. Encrypt data if encryption key/nonce
    has been cached for the given privacy request id

    :param resp_format: str, should be one of ResponseFormat
    :param data: Dict
    :param request_id: str, The privacy request id
    """
    logger.debug("Writing data to in-memory buffer")

    if resp_format == ResponseFormat.json.value:
        json_str = json.dumps(data, indent=2, default=storage_json_encoder)
        return BytesIO(
            encrypt_access_request_results(json_str, privacy_request.id).encode(
                CONFIG.security.encoding
            )
        )

    if resp_format == ResponseFormat.csv.value:
        zipped_csvs = BytesIO()
        with zipfile.ZipFile(zipped_csvs, "w") as f:
            for key in data:
                df = pd.json_normalize(data[key])
                buffer = BytesIO()
                df.to_csv(buffer, index=False, encoding=CONFIG.security.encoding)
                buffer.seek(0)
                f.writestr(
                    f"{key}.csv",
                    encrypt_access_request_results(
                        buffer.getvalue(), privacy_request.id
                    ),
                )

        zipped_csvs.seek(0)
        return zipped_csvs

    if resp_format == ResponseFormat.html.value:
        return DsrReportBuilder(
            privacy_request=privacy_request,
            dsr_data=data,
        ).generate()

    raise NotImplementedError(f"No handling for response format {resp_format}.")


def upload_to_s3(  # pylint: disable=R0913
    storage_secrets: Dict[StorageSecrets, Any],
    data: Dict,
    bucket_name: str,
    file_key: str,
    resp_format: str,
    privacy_request: Optional[PrivacyRequest],
    document: Optional[BytesIO],
    auth_method: str,
) -> Optional[AnyHttpUrlString]:
    """Uploads arbitrary data to s3 returned from an access request"""
    logger.info("Starting S3 Upload of {}", file_key)

    if privacy_request is None and document is not None:
        return generic_upload_to_s3(
            storage_secrets, bucket_name, file_key, auth_method, document
        )

    if privacy_request is None:
        raise ValueError("Privacy request must be provided")

    try:
        s3_client = get_s3_client(
            auth_method,
            storage_secrets,
            assume_role_arn=CONFIG.credentials["storage"].get("aws_s3_assume_role_arn"),
        )

        # handles file chunking
        try:
            s3_client.upload_fileobj(
                Fileobj=write_to_in_memory_buffer(resp_format, data, privacy_request),
                Bucket=bucket_name,
                Key=file_key,
            )
        except Exception as e:
            logger.error("Encountered error while uploading s3 object: {}", e)
            raise e

        presigned_url: AnyHttpUrlString = create_presigned_url_for_s3(
            s3_client, bucket_name, file_key
        )

        return presigned_url
    except ClientError as e:
        logger.error(
            "Encountered error while uploading and generating link for s3 object: {}", e
        )
        raise e
    except ParamValidationError as e:
        raise ValueError(f"The parameters you provided are incorrect: {e}")


def upload_to_gcs(
    storage_secrets: Dict,
    data: Dict,
    bucket_name: str,
    file_key: str,
    resp_format: str,
    privacy_request: PrivacyRequest,
    auth_method: str,
) -> str:
    """Uploads access request data to a Google Cloud Storage bucket"""
    logger.info("Starting Google Cloud Storage upload of {}", file_key)

    try:
        storage_client = get_gcs_client(auth_method, storage_secrets)
        bucket = storage_client.bucket(bucket_name)

        blob = bucket.blob(file_key)
        in_memory_file = write_to_in_memory_buffer(resp_format, data, privacy_request)
        content_type = {
            ResponseFormat.json.value: "application/json",
            ResponseFormat.csv.value: "application/zip",
            ResponseFormat.html.value: "application/zip",
        }
        blob.upload_from_string(
            in_memory_file.getvalue(), content_type=content_type[resp_format]
        )

        logger.info("File {} uploaded to {}", file_key, blob.public_url)

        presigned_url = blob.generate_signed_url(
            version="v4",
            expiration=CONFIG.security.subject_request_download_link_ttl_seconds,
            method="GET",
        )
        return presigned_url
    except Exception as e:
        logger.error(
            "Encountered error while uploading and generating link for Google Cloud Storage object: {}",
            e,
        )
        raise e


def upload_to_local(
    data: Dict,
    file_key: str,
    privacy_request: PrivacyRequest,
    resp_format: str = ResponseFormat.json.value,
) -> str:
    """Uploads access request data to a local folder - for testing/demo purposes only"""
    get_local_filename(file_key)

    filename = f"{LOCAL_FIDES_UPLOAD_DIRECTORY}/{file_key}"
    in_memory_file = write_to_in_memory_buffer(resp_format, data, privacy_request)

    with open(filename, "wb") as file:
        file.write(in_memory_file.getvalue())

    return "your local fides_uploads folder"



from typing import Any, Optional

from loguru import logger

from fides.api.models.messaging import MessagingConfig
from fides.api.schemas.messaging.messaging import (
    MessagingServiceDetailsAWS_SES,
    MessagingServiceSecretsAWS_SES,
)
from fides.api.schemas.storage.storage import StorageSecrets
from fides.api.util.aws_util import get_aws_session


class SESClient:
    """
    Class used to loosely type the AWS SES client.
    """

    def get_identity_verification_attributes(  # type: ignore[empty-body]
        self, Identities: list[str]
    ) -> dict[str, dict[str, dict[str, str]]]:
        """
        Returns a dict of the form:
        {
            "VerificationAttributes": {
                "email@example.com": {
                    "VerificationStatus": "Success"
                },
                "domain.com" : {
                    "VerificationStatus": "Success"
                }
            }
        }
        """

    def send_email(
        self,
        Source: str,
        Destination: dict[str, list[str]],
        Message: dict[str, Any],
    ) -> None:
        pass

    def get_template(self, TemplateName: str) -> Any:
        pass

    def create_template(self, Template: dict) -> None:
        pass


class AWS_SESException(Exception):
    pass


class AWS_SES_Service:
    """
    Service class to wrap interactions with AWS SES.
    """

    def __init__(self, messaging_config: MessagingConfig):
        """
        Instantiate AWS_SES_Service with a messaging config.
        """
        self.messaging_config_details = MessagingServiceDetailsAWS_SES.model_validate(
            messaging_config.details
        )
        self.messaging_config_secrets = MessagingServiceSecretsAWS_SES.model_validate(
            messaging_config.secrets
        )
        self._ses_client = None

    def get_ses_client(self) -> SESClient:
        """
        Returns the AWS SES client and stores it in the _ses_client variable.
        If the _ses_client variable is already set, we just return its value.
        """
        if self._ses_client is not None:
            return self._ses_client

        storage_secrets = {
            StorageSecrets.AWS_ACCESS_KEY_ID.value: self.messaging_config_secrets.aws_access_key_id,
            StorageSecrets.AWS_SECRET_ACCESS_KEY.value: self.messaging_config_secrets.aws_secret_access_key,
            "region_name": self.messaging_config_details.aws_region,
        }

        aws_session = get_aws_session(
            auth_method=self.messaging_config_secrets.auth_method.value,
            storage_secrets=storage_secrets,  # type: ignore[arg-type]
            assume_role_arn=self.messaging_config_secrets.aws_assume_role_arn,
        )
        aws_ses_client = aws_session.client(
            "ses", region_name=self.messaging_config_details.aws_region
        )

        self._ses_client = aws_ses_client
        return aws_ses_client

    def validate_email_and_domain_status(self) -> None:
        """
        Validate that either the email or domain (or both) are verified in SES.
        """
        email = self.messaging_config_details.email_from
        domain = self.messaging_config_details.domain
        identities = list(filter(None, (email, domain)))

        # Defensive code just in case, there should always be email_from or domain
        if not identities:
            raise AWS_SESException(
                "No identity (email_from or domain) configured for SES validation."
            )

        ses_client = self.get_ses_client()
        response = ses_client.get_identity_verification_attributes(
            Identities=identities
        )
        attributes = response.get("VerificationAttributes", {})

        for identity in identities:
            status = attributes.get(identity, {}).get("VerificationStatus")
            if status != "Success":
                logger.error(f"{identity} is not verified in SES.")
                raise AWS_SESException(f"{identity} is not verified in SES.")

    def send_email(
        self,
        to: str,
        subject: str,
        body: str,
    ) -> None:
        """
        Send an email using AWS SES.
        Either `email_from` or `domain` must be verified in SES.
        """
        self.validate_email_and_domain_status()
        ses_client = self.get_ses_client()

        from_address = self.messaging_config_details.email_from
        if not from_address:
            # If there is no email_from, there is a domain, and the domain was verified against SES.
            # When you verify a domain identity, you can send email from any subdomain or email address of the
            # verified domain without having to verify each one individually.
            from_address = f"noreply@{self.messaging_config_details.domain}"

        ses_client.send_email(
            Source=from_address,
            Destination={"ToAddresses": [to.strip()]},
            Message={
                "Subject": {"Data": subject},
                "Body": {"Html": {"Data": body}},
            },
        )

    def get_email_template(
        self, template_name: str
    ) -> Optional[Any]:  # pragma: no cover
        """
        Get an email template from AWS SES.
        """
        ses_client = self.get_ses_client()

        response = ses_client.get_template(TemplateName=template_name)
        return response["Template"]

    def _create_email_template(
        self, template_name: str, subject: str, text: str, html: str
    ) -> None:  # pragma: no cover
        """
        Programmatically creates an email template. This method is not called anywhere in Fides,
        because we don't create templates ourselves, but is here as a developer utility because
        AWS SES doesn't allow creating templates via the AWS Console / UI.
        """
        # Example Template
        # <img src="https://cdn.ethyca.com/fides-demo/emailheader.png" alt="Ethyca">
        # <p>Hi there,</p>
        # <span>{{fides_email_body}}</span>
        # <p>Thank you,</p>
        # <p>The Cookie House Privacy Team</p>

        ses_client = self.get_ses_client()

        ses_client.create_template(
            Template={
                "TemplateName": template_name,
                "SubjectPart": subject,
                "TextPart": text,
                "HtmlPart": html,
            }
        )



"""adds UserRegistration table

Revision ID: 8f1a19465239
Revises: 6b9885e68cbb
Create Date: 2022-10-29 00:11:51.449359

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8f1a19465239"
down_revision = "6b9885e68cbb"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "userregistration",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("user_email", sa.String(), nullable=True),
        sa.Column("user_organization", sa.String(), nullable=True),
        sa.Column("analytics_id", sa.String(), nullable=False),
        sa.Column("opt_in", sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("analytics_id"),
    )
    op.create_index(
        op.f("ix_userregistration_id"), "userregistration", ["id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_userregistration_id"), table_name="userregistration")
    op.drop_table("userregistration")
    # ### end Alembic commands ###



"""add legal basis dimension

We've previously been saving preferences against purposes, vendors, and systems, but we
need to instead save preferences against a purpose/vendor AND a legal basis.

Revision ID: 81886da90395
Revises: 9b98aba5bba8
Create Date: 2023-09-30 17:39:46.251444

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "81886da90395"
down_revision = "9b98aba5bba8"
branch_labels = None
depends_on = None


def upgrade():
    # Add six new columns to currentprivacypreference table for purpose/system/vendor x 2 legal bases
    op.add_column(
        "currentprivacypreference",
        sa.Column("purpose_consent", sa.Integer(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("purpose_legitimate_interests", sa.Integer(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("vendor_consent", sa.String(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("vendor_legitimate_interests", sa.String(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("system_consent", sa.String(), nullable=True),
    )
    op.add_column(
        "currentprivacypreference",
        sa.Column("system_legitimate_interests", sa.String(), nullable=True),
    )
    # Add 12 unique constraints - 3 new columns x 2 legal bases 2 x 2 identity types
    op.create_unique_constraint(
        "fides_user_device_identity_purpose_consent",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "purpose_consent"],
    )

    op.create_unique_constraint(
        "fides_user_device_identity_purpose_leg_interests",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "purpose_legitimate_interests"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_system_consent",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "system_consent"],
    )

    op.create_unique_constraint(
        "fides_user_device_identity_system_leg_interests",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "system_legitimate_interests"],
    )
    op.create_unique_constraint(
        "fides_user_device_identity_vendor_consent",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "vendor_consent"],
    )

    op.create_unique_constraint(
        "fides_user_device_identity_vendor_leg_interests",
        "currentprivacypreference",
        ["fides_user_device_provided_identity_id", "vendor_legitimate_interests"],
    )
    op.create_unique_constraint(
        "identity_purpose_consent",
        "currentprivacypreference",
        ["provided_identity_id", "purpose_consent"],
    )

    op.create_unique_constraint(
        "identity_purpose_leg_interests",
        "currentprivacypreference",
        ["provided_identity_id", "purpose_legitimate_interests"],
    )

    op.create_unique_constraint(
        "identity_system_consent",
        "currentprivacypreference",
        ["provided_identity_id", "system_consent"],
    )
    op.create_unique_constraint(
        "identity_system_leg_interests",
        "currentprivacypreference",
        ["provided_identity_id", "system_legitimate_interests"],
    )

    op.create_unique_constraint(
        "identity_vendor_consent",
        "currentprivacypreference",
        ["provided_identity_id", "vendor_consent"],
    )
    op.create_unique_constraint(
        "identity_vendor_leg_interests",
        "currentprivacypreference",
        ["provided_identity_id", "vendor_legitimate_interests"],
    )

    # Add six new indices - for purpose/vendor/system columns x 2 legal bases
    op.create_index(
        op.f("ix_currentprivacypreference_purpose_consent"),
        "currentprivacypreference",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_purpose_legitimate_interests"),
        "currentprivacypreference",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_system_consent"),
        "currentprivacypreference",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_system_legitimate_interests"),
        "currentprivacypreference",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_vendor_consent"),
        "currentprivacypreference",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_currentprivacypreference_vendor_legitimate_interests"),
        "currentprivacypreference",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    # Add six new columns to lastservednotice for purpose, vendor, and consent x 2 legal bases
    op.add_column(
        "lastservednotice", sa.Column("purpose_consent", sa.Integer(), nullable=True)
    )
    op.add_column(
        "lastservednotice",
        sa.Column("purpose_legitimate_interests", sa.Integer(), nullable=True),
    )
    op.add_column(
        "lastservednotice", sa.Column("vendor_consent", sa.String(), nullable=True)
    )
    op.add_column(
        "lastservednotice",
        sa.Column("vendor_legitimate_interests", sa.String(), nullable=True),
    )
    op.add_column(
        "lastservednotice", sa.Column("system_consent", sa.String(), nullable=True)
    )
    op.add_column(
        "lastservednotice",
        sa.Column("system_legitimate_interests", sa.String(), nullable=True),
    )
    # Add six new indices to lastservednotice for purpose, vendor, and consent x 2 legal bases

    op.create_index(
        op.f("ix_lastservednotice_purpose_consent"),
        "lastservednotice",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_purpose_legitimate_interests"),
        "lastservednotice",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_system_consent"),
        "lastservednotice",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_system_legitimate_interests"),
        "lastservednotice",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_vendor_consent"),
        "lastservednotice",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_lastservednotice_vendor_legitimate_interests"),
        "lastservednotice",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    # Add 12 unique constraints to lastserved notice - purpose/vendor/system x 2 legal bases x 2 identity types
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_purpose_consent",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "purpose_consent"],
    )

    op.create_unique_constraint(
        "last_served_fides_user_device_identity_purpose_leg_interests",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "purpose_legitimate_interests"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_system_consent",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "system_consent"],
    )

    op.create_unique_constraint(
        "last_served_fides_user_device_identity_system_leg_interests",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "system_legitimate_interests"],
    )
    op.create_unique_constraint(
        "last_served_fides_user_device_identity_vendor_consent",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "vendor_consent"],
    )

    op.create_unique_constraint(
        "last_served_fides_user_device_identity_vendor_leg_interests",
        "lastservednotice",
        ["fides_user_device_provided_identity_id", "vendor_legitimate_interests"],
    )
    op.create_unique_constraint(
        "last_served_identity_purpose_consent",
        "lastservednotice",
        ["provided_identity_id", "purpose_consent"],
    )
    op.create_unique_constraint(
        "last_served_identity_purpose_legitimate_interests",
        "lastservednotice",
        ["provided_identity_id", "purpose_legitimate_interests"],
    )

    op.create_unique_constraint(
        "last_served_identity_system_consent",
        "lastservednotice",
        ["provided_identity_id", "system_consent"],
    )
    op.create_unique_constraint(
        "last_served_identity_system_leg_interests",
        "lastservednotice",
        ["provided_identity_id", "system_legitimate_interests"],
    )

    op.create_unique_constraint(
        "last_served_identity_vendor_consent",
        "lastservednotice",
        ["provided_identity_id", "vendor_consent"],
    )
    op.create_unique_constraint(
        "last_served_identity_vendor_leg_interests",
        "lastservednotice",
        ["provided_identity_id", "vendor_legitimate_interests"],
    )

    # Add six new columns to privacypreferencehistory - purpose/vendor/system x 2 legal bases
    op.add_column(
        "privacypreferencehistory",
        sa.Column("purpose_consent", sa.Integer(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("purpose_legitimate_interests", sa.Integer(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("vendor_consent", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("vendor_legitimate_interests", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("system_consent", sa.String(), nullable=True),
    )
    op.add_column(
        "privacypreferencehistory",
        sa.Column("system_legitimate_interests", sa.String(), nullable=True),
    )
    # Add six indices to privacypreference history - purpose/vendor/system x 2 legal bases
    op.create_index(
        op.f("ix_privacypreferencehistory_purpose_consent"),
        "privacypreferencehistory",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_purpose_legitimate_interests"),
        "privacypreferencehistory",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_system_consent"),
        "privacypreferencehistory",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_system_legitimate_interests"),
        "privacypreferencehistory",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_vendor_consent"),
        "privacypreferencehistory",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacypreferencehistory_vendor_legitimate_interests"),
        "privacypreferencehistory",
        ["vendor_legitimate_interests"],
        unique=False,
    )
    # Add six new columns to servednoticehistory
    op.add_column(
        "servednoticehistory", sa.Column("purpose_consent", sa.Integer(), nullable=True)
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("purpose_legitimate_interests", sa.Integer(), nullable=True),
    )
    op.add_column(
        "servednoticehistory", sa.Column("vendor_consent", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("vendor_legitimate_interests", sa.String(), nullable=True),
    )
    op.add_column(
        "servednoticehistory", sa.Column("system_consent", sa.String(), nullable=True)
    )
    op.add_column(
        "servednoticehistory",
        sa.Column("system_legitimate_interests", sa.String(), nullable=True),
    )
    # Add six new indices to servednoticehistory
    op.create_index(
        op.f("ix_servednoticehistory_purpose_consent"),
        "servednoticehistory",
        ["purpose_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_purpose_legitimate_interests"),
        "servednoticehistory",
        ["purpose_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_system_consent"),
        "servednoticehistory",
        ["system_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_system_legitimate_interests"),
        "servednoticehistory",
        ["system_legitimate_interests"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_vendor_consent"),
        "servednoticehistory",
        ["vendor_consent"],
        unique=False,
    )
    op.create_index(
        op.f("ix_servednoticehistory_vendor_legitimate_interests"),
        "servednoticehistory",
        ["vendor_legitimate_interests"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_servednoticehistory_vendor_legitimate_interests"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_vendor_consent"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_system_legitimate_interests"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_system_consent"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_servednoticehistory_purpose_legitimate_interests"),
        table_name="servednoticehistory",
    )
    op.drop_index(
        op.f("ix_servednoticehistory_purpose_consent"), table_name="servednoticehistory"
    )
    op.drop_column("servednoticehistory", "system_legitimate_interests")
    op.drop_column("servednoticehistory", "system_consent")
    op.drop_column("servednoticehistory", "vendor_legitimate_interests")
    op.drop_column("servednoticehistory", "vendor_consent")
    op.drop_column("servednoticehistory", "purpose_legitimate_interests")
    op.drop_column("servednoticehistory", "purpose_consent")
    op.drop_index(
        op.f("ix_privacypreferencehistory_vendor_legitimate_interests"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_vendor_consent"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_system_legitimate_interests"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_system_consent"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_purpose_legitimate_interests"),
        table_name="privacypreferencehistory",
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_purpose_consent"),
        table_name="privacypreferencehistory",
    )
    op.drop_column("privacypreferencehistory", "system_legitimate_interests")
    op.drop_column("privacypreferencehistory", "system_consent")
    op.drop_column("privacypreferencehistory", "vendor_legitimate_interests")
    op.drop_column("privacypreferencehistory", "vendor_consent")
    op.drop_column("privacypreferencehistory", "purpose_legitimate_interests")
    op.drop_column("privacypreferencehistory", "purpose_consent")
    op.drop_constraint(
        "last_served_identity_vendor_consent", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_vendor_leg_interests", "lastservednotice", type_="unique"
    )

    op.drop_constraint(
        "last_served_identity_system_consent", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_system_leg_interests", "lastservednotice", type_="unique"
    )

    op.drop_constraint(
        "last_served_identity_purpose_consent", "lastservednotice", type_="unique"
    )
    op.drop_constraint(
        "last_served_identity_purpose_legitimate_interests",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_vendor_consent",
        "lastservednotice",
        type_="unique",
    )

    op.drop_constraint(
        "last_served_fides_user_device_identity_vendor_leg_interests",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_system_consent",
        "lastservednotice",
        type_="unique",
    )

    op.drop_constraint(
        "last_served_fides_user_device_identity_system_leg_interests",
        "lastservednotice",
        type_="unique",
    )
    op.drop_constraint(
        "last_served_fides_user_device_identity_purpose_consent",
        "lastservednotice",
        type_="unique",
    )

    op.drop_constraint(
        "last_served_fides_user_device_identity_purpose_leg_interests",
        "lastservednotice",
        type_="unique",
    )
    op.drop_index(
        op.f("ix_lastservednotice_vendor_legitimate_interests"),
        table_name="lastservednotice",
    )
    op.drop_index(
        op.f("ix_lastservednotice_vendor_consent"), table_name="lastservednotice"
    )
    op.drop_index(
        op.f("ix_lastservednotice_system_legitimate_interests"),
        table_name="lastservednotice",
    )
    op.drop_index(
        op.f("ix_lastservednotice_system_consent"), table_name="lastservednotice"
    )
    op.drop_index(
        op.f("ix_lastservednotice_purpose_legitimate_interests"),
        table_name="lastservednotice",
    )
    op.drop_index(
        op.f("ix_lastservednotice_purpose_consent"), table_name="lastservednotice"
    )
    op.drop_column("lastservednotice", "system_legitimate_interests")
    op.drop_column("lastservednotice", "system_consent")
    op.drop_column("lastservednotice", "vendor_legitimate_interests")
    op.drop_column("lastservednotice", "vendor_consent")
    op.drop_column("lastservednotice", "purpose_legitimate_interests")
    op.drop_column("lastservednotice", "purpose_consent")
    op.drop_index(
        op.f("ix_currentprivacypreference_vendor_legitimate_interests"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_vendor_consent"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_system_legitimate_interests"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_system_consent"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_purpose_legitimate_interests"),
        table_name="currentprivacypreference",
    )
    op.drop_index(
        op.f("ix_currentprivacypreference_purpose_consent"),
        table_name="currentprivacypreference",
    )
    op.drop_constraint(
        "identity_vendor_consent", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "identity_vendor_leg_interests", "currentprivacypreference", type_="unique"
    )

    op.drop_constraint(
        "identity_system_consent", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "identity_system_leg_interests", "currentprivacypreference", type_="unique"
    )

    op.drop_constraint(
        "identity_purpose_consent", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "identity_purpose_leg_interests", "currentprivacypreference", type_="unique"
    )
    op.drop_constraint(
        "fides_user_device_identity_vendor_consent",
        "currentprivacypreference",
        type_="unique",
    )

    op.drop_constraint(
        "fides_user_device_identity_vendor_leg_interests",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_constraint(
        "fides_user_device_identity_system_consent",
        "currentprivacypreference",
        type_="unique",
    )

    op.drop_constraint(
        "fides_user_device_identity_system_leg_interests",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_constraint(
        "fides_user_device_identity_purpose_consent",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_constraint(
        "fides_user_device_identity_purpose_leg_interests",
        "currentprivacypreference",
        type_="unique",
    )
    op.drop_column("currentprivacypreference", "system_legitimate_interests")
    op.drop_column("currentprivacypreference", "system_consent")
    op.drop_column("currentprivacypreference", "vendor_legitimate_interests")
    op.drop_column("currentprivacypreference", "vendor_consent")
    op.drop_column("currentprivacypreference", "purpose_legitimate_interests")
    op.drop_column("currentprivacypreference", "purpose_consent")



from fideslang.validation import AnyHttpUrlString
from pydantic import BaseModel

from fides.api.schemas.base_class import NoValidationSchema


class WebsiteSchema(BaseModel):
    """Schema to validate the secrets needed for a generic website connector"""

    url: AnyHttpUrlString


class WebsiteDocsScehma(WebsiteSchema, NoValidationSchema):
    """Website Secrets Schema for API Docs"""
